%******************************************************************************;
%* SAS macro _Append_ vs. 0.2.1 , 17 Oct 2013, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%*                                                                            *;
   %MACRO   _Append_   /* PROC APPEND substitute adding new variables as well */
/* -------------------------------------------------------------------------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  Base    = _LAST_    /* Dataset name to append Data to                    R */
,  Data    = _LAST_    /* Dataset name of Data to append to Base            R */
,  Out     =           /* empty: same as Base, else different dataset       O */
,  CalledBy=_Append_   /* Macro name of macro or program for err0r report   O */
,  Version = 0.2.1     /* forced version specification for version control  O */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  ForMyUse=           /* argument for the author's use only, do not use it ! */
)/ DES     =             'PROC APPEND alternative new variables'  /* STORE */  ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : _Append_.SAS                                                  *;
%*                                                                            *;
%* Purpose    : PROC APPEND refuses to add new variables to an already        *;
%*              existing dataset. This macro does the same as PROC APPEND,    *;
%*              and also supports adding new variables to a dataset.          *;
%*              The PROC APPEND option FORCE is not applicable as different   *;
%*              character variable lengths are always allowed and the longest *;
%*              length is taken as the output length, no loss of information. *;
%*                                                                            *;
%* Environment: C A L L   E X E C U T E - p r o o f (no dependent macro code) *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 17 Oct 2013                                                   *;
%* Version    : 0.2.1                                                         *;
%*                                                                            *;
%* SAS version: 9.1.3 and WXP                                                 *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Examples                                                                   *;
%* --------                                                                   *;
%* %_Append_ (Base=One, Data=Two) /* appending Two to One, if no One: Two */  *;
%* %_Append_ (Base=One, Data=Two, Out=Appended, CalledBy=Program) /* no ; */  *;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
%* = internal: - [none]                                                       *;
%* = attached: - [none]                                                       *;
%* = external: - [none]                                                       *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of _Append_                                                        *;
%* ===================                                                        *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0.0    28 Aug 2009  initial working release, does not fix lengths        *;
%*                       strategies:                                          *;
%*                         Standard (PROC APPEND with its drawbacks),         *;
%*                         Bad (data step, fails if &Base not exists),        *;
%*                         Regular (data step, fails from CALL EXECUTE),      *;
%*                         Delayed (data step, fails from CALL EXECUTE),      *;
%*                         Better (data step, no maximum length adaptation),  *;
%*                         AsGood (data step, no maximum length adaptation),  *;
%* 0.0.1    26 Oct 2009  added BEST algorithm with PROC CONTENTS and SAS code *;
%*                       added Out= argument for alternative output dataset   *;
%*           2 Nov 2009  added forced alpha sorting of PROC CONTENTS OUTput   *;
%* 0.0.2     2 Nov 2009  removed experimental code (alternative algorithms),  *;
%*                       added 'CALL EXECUTE - proof' certificate             *;
%*                       added _LAST_ support and correct SYSLAST spec at end,*;
%*                          (unsure about correct functioning of _LAST_ and   *;
%*                          SYSLAST from CALL EXECUTE, avoid it at the time)  *;
%*           3 nov 2009  suppressed log header with default CalledBy argument *;
%* 0.1.0    20 Jul 2010  a lot of changes, improvements, potential bug fixes: *;
%*                       removed some outdated, commented code, added comment,*;
%*                       fixed potential bugs with same named variables with  *;
%*                         different case in both datasets, changed to upcase,*;
%*                       forced new variables from &Data to be added after    *;
%*                         all variables of &Base (by manipulating Varnum),   *;
%*          21 Jul 2010  forced abortion in case of discrepant variable types,*;
%*                       completely revised the algorithm and code to deduce  *;
%*                         the optimal output lengths, labels, formats and    *;
%*                         informats. The previous code was not entirely      *;
%*                         correct and optimal. This may fix a few bugs.      *;
%*                         - Variables now are defined by a LENGTH statement, *;
%*                         - Firstly encountered variable labels in the       *;
%*                           datasets are preserved, other ones are discarded,*;
%*                         - Maximum character lengths determine (in)formats, *;
%*                         - Firstly encountered numerical (in)format names   *;
%*                           determine the (in)formats, these are copied,     *;
%*                         - If there are no numerical (in)format names then  *;
%*                           the (in)formats are deduced from maximum widths  *;
%*                           and maximum number of decimals in the w.d spec.  *;
%*                       changed work dataset names to macro variables with   *;
%*                         quite uncommon, unique names, starting with an _,  *;
%*                       added deletion of work datasets when finished        *;
%* 0.1.1    17 Mar 2011  fixed bug with uninitialised {dInform} informat      *;
%* 0.2.0    11 Oct 2013  added support for non-matching variable types of     *;
%*                         same variables by converting numerical variables   *;
%*                         to character variables (as far as necessary) using *;
%*                         external (auxiliary) macro Num2Char                *;
%* 0.2.1    17 Oct 2013  added support for SAS name literal variable names    *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS NOMPRINT NOMERROR NOSERROR NOMLOGIC NOSYMBOLGEN NOMACROGEN;
%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* remove comment chars for testing ;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN %DO;
    %GOTO Finish;
  %END;

%IF (&Debug EQ 1 OR &Debug GE 11) %THEN %DO;
  OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* LOCAL macro variables declarations ;
%*____________________________________;

 %LOCAL    /* Define internal macro variables explicitely local */
  MacName  /* Name of macro: RemBugNL */
  MacroVs  /* Version number major.minor.patch of macro: 0.2.1 */
  MacDate  /* Date of this version of macro: 17 Oct 2013 */
  Err      /* text 'E R R O R' as consecutive characters */
  Warn     /* text 'W A R N I N G' as consecutive characters */
  BaseCnts /* substitute macro variable name for Contents of &Base */
  DataCnts /* substitute macro variable name for Contents of &Data */
  MetaData /* substitute macro variable name for merged contents */
  EmptyPDV /* substitute macro variable name for variable structure */
  MxVarnum /* number of variables in &Base */
  BaseConv /* possibly numeric to character converted Base dataset */
  DataConv /* possibly numeric to character converted Data dataset */
  bVarList /* numeric variable list in &Base to convert to character */
  dVarList /* numeric variable list in &Data to convert to character */
  Checking /* checking existence of macro (Num2Char) */
  ;

  %LET MacName  = _Append_;
  %LET MacroVs  = 0.2.1;
  %LET MacDate  = %STR(17 Oct 2013);
  %LET Err      = %SYSFUNC(COMPRESS(E R R O R, %STR( )));
  %LET Warn     = %SYSFUNC(COMPRESS(W A R N I N G, %STR( )));
  %LET BaseCnts = __Base__;
  %LET DataCnts = __Data__;
  %LET MetaData = __Meta__;
  %LET EmptyPDV = __PDV__;

%* Macro header in LOG file, dependent on value of CalledBy;
 %IF ( %UPCASE(&CalledBy) NE %UPCASE(&MacName) ) %THEN
 %DO;
  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT | Macro &MacName, vs. &MacroVs, by Jim Groeneveld, &MacDate |;
  %PUT |    PROC APPEND substitute adding new variables as well    |;
  %PUT |___________________________________________________________|;
  %PUT;

%* Feedback of specified arguments;
  %PUT &MacName: Base = &Base;
  %PUT &MacName: Data = &Data;
  %PUT &MacName: Out = &Out;
  %PUT &MacName: CalledBy = &CalledBy;
  %PUT &MacName: Version  = &Version ;
  %*PUT &MacName: ForMyUse = &ForMyUse;
 %END;

  %IF (&Version NE AND &Version NE &MacroVs) %THEN %DO;
    %PUT *** &MacName *** &ERR: Specified version &Version does not match &MacroVs,;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;

%*~~~~~~~~~~~~~~~~~;
%* Parameter checks;
%*_________________;

%* Base empty;
  %IF (%LENGTH(&Base) EQ 0) %THEN %DO;
    %PUT *** &MacName *** &ERR: Base value illegally is empty,;
    %PUT .                  macro &MacName will abort;
    %GOTO Finish;
  %END;
%* Data empty;
  %IF (%LENGTH(&Data) EQ 0) %THEN %DO;
    %PUT *** &MacName *** &ERR: Data value illegally is empty,;
    %PUT .                  macro &MacName will abort;
    %GOTO Finish;
  %END;
%* Base and Data different;
  %ELSE %IF (%QUOTE(&Base) EQ %QUOTE(&Data)) %THEN %DO;
    %PUT *** &MacName *** &ERR: Base and Data values illegally are identical,;
    %PUT .                  macro &MacName will abort;
    %GOTO Finish;
  %END;

%* Process keyword _LAST_ with both Base and Data;
  %IF (%UPCASE(&Base) EQ _LAST_) %THEN %LET Base = &SYSLAST;
  %IF (%UPCASE(&Data) EQ _LAST_) %THEN %LET Data = &SYSLAST;

%*~~~~~~;
%* Core ;
%*______;

  %IF (%LENGTH(&Out) EQ 0) %THEN %LET Out=&Base;
  DATA _NULL_;
* If &Base does not exist then its PDV is created from &Data at runtime;
    IF (NOT (EXIST("&Base"))) THEN /* get PDV, attributes of &Data */
      CALL EXECUTE ("DATA &Base; IF 0 THEN SET &Data; STOP; RUN;");
  RUN;

* &Base now exists;
* Copy &Base and &Data names to &BaseConv and &DataConv;
  %LET BaseConv = &Base;
  %LET DataConv = &Data;

%Repeat: %* Also repeat partial code below after type conversion;
  %IF ((&BaseConv EQ &Base AND &DataConv EQ &Data) /* 1st time */
    OR  &BaseConv NE &Base /* converted &Base */) %THEN %DO;
    PROC CONTENTS DATA=&BaseConv
                   OUT=&BaseCnts (KEEP=Name Type Length Label VarNum
                     Format FormatL FormatD InFormat InFormL InFormD)
                   NOPRINT;
    RUN;

* Make variable names all upper case (they already were in a previous SAS version);
    DATA &BaseCnts; SET &BaseCnts; Name = UPCASE(Name); RUN;
    PROC SORT DATA=&BaseCnts; BY Name; RUN; * now original case independent;
  %END; * producing contents of &Base or &BaseConv;

  %IF ((&BaseConv EQ &Base AND &DataConv EQ &Data) /* 1st time */
    OR  &DataConv NE &Data /* converted &Data */) %THEN %DO;
    PROC CONTENTS DATA=&DataConv
                   OUT=&DataCnts (KEEP=Name Type Length Label VarNum
                     Format FormatL FormatD InFormat InFormL InFormD
                     RENAME=(Type=dType Length=dLength Label=dLabel /*VarNum=dVarNum*/
                             Format=dFormat FormatL=dFormatL FormatD=dFormatD
                             Informat=dInform InFormL=dInFormL InFormD=dInFormD))
                   NOPRINT;
    RUN;

* Make variable names all upper case (they already were in a previous SAS version);
    DATA &DataCnts; SET &DataCnts; Name = UPCASE(Name); RUN;
    PROC SORT DATA=&DataCnts; BY Name; RUN; * now original case independent;
  %END; * producing contents of &Data or &DataConv;

* Determine number of observations in &BaseCnts = number of variables in &Base;
  %IF (&BaseConv EQ &Base AND &DataConv EQ &Data) /* 1st time */ %THEN %DO;
    DATA _NULL_;
      IF (0) THEN SET &BaseCnts NOBS=MxVarnum;
      CALL SYMPUT ('MxVarnum', PUT (MxVarnum, 8.));
      STOP;
    RUN;
  %END; %* determination of number of observations in BaseCnts, 1st time only;

  %LET bVarList = ; * initialise numerical variable list from &Base to convert;
  %LET dVarList = ; * initialise numerical variable list from &Data to convert;

  DATA &MetaData;
    MERGE &DataCnts &BaseCnts (IN=InBase) ; BY Name; * reverse order of datasets;
    * to conserve Varnum of &BaseCnts dataset in case it is in both datasets.;
    RETAIN MxVarNum &MxVarNum;
    IF (InBase) THEN MxVarNum + (-1);
    ELSE Varnum + MxVarNum; * increase Varnum of;
    * records in &DataCnts only by number of records in &BaseCnts;
    * This does not yet generate the proper sequence (there are holes);
    * but it forces the correct order anyway when sorting;

    IF (Type NE dType AND NOT MISSING(Type) AND NOT MISSING(dType)) THEN DO;
* Store numeric variable names of &Base and/or &Data in macro variable(s);
      IF (Type EQ 1) THEN /* &Base contains numeric variable to convert */
        CALL EXECUTE ('%LET bVarList = &bVarList ' || TRIM(NLITERAL(Name)) || ';');
      ELSE /* dType=1, &Data contains numeric variable to convert */
        CALL EXECUTE ('%LET dVarList = &dVarList ' || TRIM(NLITERAL(Name)) || ';');
    END;
  RUN;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* If any &Base and/or &Data dataset contain same named variables of ;
* different Type the numeric ones have to be converted to character.;
* Subsequently a part of the previous processing has to be repeated ;
* once to be processed further without non-matching variable types. ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

  %IF (%LENGTH(%SUPERQ(bVarList)) OR %LENGTH(%SUPERQ(dVarList))) %THEN %DO;
%* Firstly check presence of macro Num2Char;
    %LET Checking = %Num2Char(Presence=%NRSTR(&MacName));
    %IF (&Checking NE Num2Char) %THEN %DO;
      %PUT *** &MacName *** &ERR: Macro Num2Char not present,;
      %PUT install it and rerun. Macro &MacName will abort!;
      %GOTO Aborting;
    %END;
%* Convert variables in &bVarList into ____Base dataset;
    %IF (%LENGTH(%SUPERQ(bVarList))) %THEN %DO;
      %LET BaseConv = ____Base; %* new, unique temporary dataset;
      %Num2Char (Data=&Base, Out=&BaseConv, Overwrit=Yes,
                 VarList=&bVarList, CalledBy=&MacName);
    %END;
%* Convert variables in &dVarList into ____Data dataset;
    %IF (%LENGTH(%SUPERQ(dVarList))) %THEN %DO;
      %LET DataConv = ____Data; %* new, unique temporary dataset;
      %Num2Char (Data=&Data, Out=&DataConv, Overwrit=Yes,
                 VarList=&dVarList, CalledBy=&MacName);
    %END;
    %GOTO Repeat; %* (this may be reprogrammed without a GOTO later);
  %END; %* of conversion and repetition of a part of the above code;

%* Continue further processing with always compatible variable types;
  PROC SORT DATA=&MetaData; BY VarNum; RUN; * conserving original order ;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Because the current strategy rebuilds a resulting dataset from scratch,  ;
* firstly building a PDV with widest character variables from merged PROC  ;
* CONTENTS OUTput and subsequently concatenating both datasets to that,the ;
* original attributes of the variables from both datasets,if any,are lost. ;
* Therefore the most important attributes, (max.) lengths and (in)formats  ;
* from the original variables in the original datasets are preserved and   ;
* combined as good as possible and added to the resulting appended dataset.;
*__________________________________________________________________________;

  DATA &MetaData;
* Process &MetaData and generate dependent data step;
    SET &MetaData END=LastRec;
* Copy dType of new variable (in &DataCnts) to Type;
    IF (NOT MISSING(dType) AND MISSING(Type)) THEN Type = dType;

* Define full (in)format length specification;
    LENGTH Fmt InFmt $50; * 49 = 32 + 8 + 1 + 8;

* Determine format name from &BaseCnts, else from &DataCnts (else empty);
*   If Format exists then leave its width and decimals, else ...;
    IF (Format EQ '') THEN DO; * if only &DataCnts_ format name;
      Format = dFormat; FormatL = dFormatL; FormatD = dFormatD;
    END;
    ELSE IF (dFormat EQ '') THEN DO; * both format names empty;
* Determine maximum format Width of numeric variable;
      FormatL = MAX (FormatL, dFormatL);
* Determine maximum number of decimals for numeric informat;
      FormatD = MAX (FormatD, dFormatD); * arbitrary;
    END;

* Determine informat name from &BaseCnts, else from &DataCnts (else empty);
*   If Informat exists then leave its width and decimals, else ...;
    IF (InFormat EQ '') THEN DO; * if only &DataCnts_ informat name;
      InFormat = dInForm; InFormL = dInFormL; InFormD = dInFormD;
    END;
    ELSE IF (dInForm EQ '') THEN DO; * both format names empty;
* Determine maximum informat width of numeric variable;
      InFormL = MAX (InFormL, dInFormL);
* Determine maximum number of decimals for numeric informat;
      InFormD = MAX (InFormD, dInFormD); * arbitrary;
    END;

* Determine maximum character variable length;
    Length = MAX (Length, dLength);

* Start creating empty dataset, just PDV;
    IF (_N_ EQ 1) THEN CALL EXECUTE ("DATA &EmptyPDV;");

* If numeric variable then just name it to force its existence in order;
    IF (Type EQ 1) THEN DO;
* Define length of numeric variable, always 8;
      CALL EXECUTE ('LENGTH ' || TRIM(NLITERAL(Name)) || ' ' || TRIM(LEFT(PUT(Length,8.))) || ';');
* Generate maximum (in)format;
      IF (FormatL) THEN
        Fmt   = TRIM(Format)   || TRIM(LEFT(PUT(FormatL,8.))) || '.'
                               || TRIM(LEFT(PUT(FormatD,8.)));
      IF (InFormL) THEN
        InFmt = TRIM(InFormat) || TRIM(LEFT(PUT(InFormL,8.))) || '.'
                               || TRIM(LEFT(PUT(InFormD,8.)));
    END;

* If character variable then define it with maximum length;
    ELSE IF (Type EQ 2) THEN DO;
      CALL EXECUTE ('LENGTH ' || TRIM(NLITERAL(Name)) || ' $' || TRIM(LEFT(PUT(Length,8.))) || ';');
* Generate maximum (in)format Width as its length;
      IF (Format EQ '')   THEN Format = '$';   * add explicit $, though redundant? ;
      IF (InFormat EQ '') THEN InFormat = '$'; * add explicit $, though redundant? ;
      Fmt   = TRIM(Format)   || TRIM(LEFT(PUT(Length,8.))) || '.'; * incl. $ char.;
      InFmt = TRIM(InFormat) || TRIM(LEFT(PUT(Length,8.))) || '.'; * incl. $ char.;
    END;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* When creating an empty PDV it neither yet contains the variable attributes, ;
* like Label, Format and Informat. These attributes have been determined      ;
* above and will be added below.                                              ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

* Copy attribute LABEL to PDV ('empty' first dataset);
    IF (NOT MISSING(Label)) THEN CALL EXECUTE
       ('LABEL ' || TRIM(NLITERAL(Name)) || '=' || QUOTE(TRIM(Label)) || ';' );
    ELSE IF (NOT MISSING(dLabel)) THEN CALL EXECUTE
       ('LABEL ' || TRIM(NLITERAL(Name)) || '=' || QUOTE(TRIM(dLabel)) || ';' );

* Copy attribute FORMAT ('empty' first dataset);
    IF (NOT MISSING(Fmt)) THEN CALL EXECUTE
       ('FORMAT ' || TRIM(NLITERAL(Name)) || ' ' || TRIM(Fmt) || ';' );
* Copy attribute INFORMAT ('empty' first dataset);
    IF (NOT MISSING(InFmt)) THEN CALL EXECUTE
       ('INFORMAT ' || TRIM(NLITERAL(Name)) || ' ' || TRIM(InFmt) || ';' );

* Finalyse creating empty dataset, just variables;
    IF (LastRec) THEN CALL EXECUTE ('STOP; RUN;');
  RUN;

* CALL EXECUTE generated code runs here creating &EmptyPDV dataset! ;

  %IF (&Debug EQ 5) %THEN %DO; %* Libname Current must be defined! ;
    libname current '.';
    data current.&MetaData; set &MetaData; run; * debug;
    data current.&EmptyPDV; set &EmptyPDV; run; * debug;
  %END;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Concatenation of lossless PDV, datasets Base and Data ;
*_______________________________________________________;

  DATA &Out; SET &EmptyPDV &BaseConv &DataConv; RUN; * Actual concatenation to PDV;

  %LET SYSLAST = &Out;

%Aborting:
  %_Delete_ (&BaseCnts &DataCnts &MetaData &EmptyPDV ____Base ____Data);

%*=========;
%Finish:

%*~~~~~~~~~~~~~;
%MEND _Append_ ;
%*_____________;
